#include "Download.h"
#include "DownloadManager.h"
#include "StringManipulation.h"
#include "QtellaSub.h"
#include "Connections.h"
#include "Push.h"
#include "Logging.h"
#include "DownloadViewItem.h"
#include "Icon.h"
#include "InterruptedDownloads.h"

#include <qsocket.h>
#include <qlistview.h>
#include <qfile.h>
#include <qdir.h>
#include <qradiobutton.h>
#include <qspinbox.h>
#include <qlineedit.h>
#include <qpixmap.h>

#include <iostream>
#include <iomanip>
#include <strstream>
#include <ctime>

Download::Download(QueryHitEntry qhe, DownloadManager *parent, QListViewItem* item, std::string dest, QListViewItem* int_item) : 
  _parent(parent), _qhe(qhe), _state(dtQueued), _content_length(qhe.size), _retries(0),
  sock(NULL), _active(false), _search_item(item), _all_read(0), _dest_file(dest), _int_item(int_item)
{
  if(dest.empty()) _dest_file = qhe.filename;

  _c = QtellaSub::getInstance();

  item->setPixmap(2, *(Icon::getInstance()->getYellowBall()));
  _c->_interrupted_downloads->setYellow(_int_item);

  _item = new DownloadViewItem(_parent->_parent->ui_listview_download);

  timeout = new QTimer();
  timer_bandwidth = new QTimer();
  timer_retry = new QTimer();
  
  if(!_qhe.firewalled())
    {
      sock = new QSocket();

      connect(sock, SIGNAL(readyRead()), this, SLOT(slotReadyRead()));
      connect(sock, SIGNAL(connectionClosed()), this, SLOT(slotConnectionClosed()));
      connect(sock, SIGNAL(error(int)), this, SLOT(slotError(int)));
      connect(sock, SIGNAL(connected()), this, SLOT(slotConnected()));
    }
      
  connect(timeout, SIGNAL(timeout()), this, SLOT(slotTimeout()));
  connect(timer_retry, SIGNAL(timeout()), this, SLOT(slotRetry()));
  connect(timer_bandwidth, SIGNAL(timeout()), this, SLOT(slotTimerBandwidth()));

  _client.erase();
  _first_start = true;
  _f = NULL;
}


void Download::run()
{
  _all_read = 0;

  setState(dtConnecting);
  _active = true;

  if(timer_retry->isActive()) timer_retry->stop();
  if(timeout->isActive()) timeout->stop();
  if(timer_bandwidth->isActive()) timer_bandwidth->stop();

  bandwidth     = 0;
  tmp_read      = 0;
  _buffer.erase();
  _header.erase();
  retry_timeout = 30;

  std::string host = _qhe.addr.strIP();


  complete_path   = _parent->_parent->ui_lineedit_downloaddirectory->text();
  if(complete_path[complete_path.size()-1] != '/') complete_path += "/";
  complete_path += _dest_file;

  incomplete_path = _parent->_parent->ui_lineedit_incompleted->text();
  if(incomplete_path[incomplete_path.size()-1] != '/') incomplete_path += "/";
  incomplete_path += _dest_file;


  QDir d(incomplete_path.substr(0, incomplete_path.find_last_of("/")).c_str());

  
  if(!d.exists())
    {
      std::string error("Directory for incomplete downloads does not exist! [");
      error += incomplete_path.substr(0, incomplete_path.find_last_of("/")) + "]";
      Logging::getInstance()->log(Logging::Error, error);
      setState(dtError);
      return;
    }

  d.setPath(complete_path.substr(0, complete_path.find_last_of("/")).c_str());

  if(!d.exists())
    {
      std::string error("Directory for completed downloads does not exist!");
      error += complete_path.substr(0, complete_path.find_last_of("/")) + "]";
      Logging::getInstance()->log(Logging::Error, error);
      setState(dtError);
      return;
    }

  std::string debug("Start downloading file. [" + std::string(incomplete_path) + "]");
  Logging::getInstance()->log(Logging::Debug, debug);

  if(_first_start)
    {
      _first_start = false;

      QFile file(complete_path.c_str());

      if( file.exists() && _parent->_parent->ui_radiobutton_e_abort->isChecked() )
	{
	  _f = new QFile(complete_path.c_str());
	  setState(dtExists);
	  return;
	}

      file.setName(incomplete_path.c_str());


      if( _parent->_parent->ui_radiobutton_a_startnew->isChecked() )
	{
	  std::string s = createFilename(incomplete_path, _qhe.size, _qhe.addr.strIP());
	  _f = new QFile(s.c_str());
	}
      else
	{
	  std::string s;

	  if( _parent->_parent->ui_radiobutton_a_contsize->isChecked() )
	    s = findFile(incomplete_path, _qhe.size);

	  if( _parent->_parent->ui_radiobutton_a_conthost->isChecked() )
	    s = findFile(incomplete_path, _qhe.addr.strIP());

	  if( _parent->_parent->ui_radiobutton_a_contsizehost->isChecked() )
	    s = findFile(incomplete_path, _qhe.size, _qhe.addr.strIP());
	      
	  if(!s.empty())
	    _f = new QFile(s.c_str());
	  else
	    {
	      std::string s = createFilename(incomplete_path, _qhe.size, _qhe.addr.strIP());
	      _f = new QFile(s.c_str());
	    }
	}

    }
  else
    if(_f->isOpen()) _f->close();

  
  if(!_f->isOpen()) _f->open(IO_Append|IO_ReadWrite|IO_Raw);
  if(!_f->isOpen())
    {
      std::string error("Cannot open file for writing! [" + std::string(_f->name()) + "]");
      Logging::getInstance()->log(Logging::Error, error);
      slotError(dtError);
      return;
    }

  if(_f->size() > 0) retry_timeout = 15;

  timeout->start(20000, true);

  if(!_qhe.firewalled())
    sock->connectToHost(_qhe.addr.strIP().c_str(), _qhe.addr.port());
  else
    {
      std::string id = _parent->_parent->connections->_id;
      Address addr( std::string(_parent->_parent->ui_lineedit_ip->text()) );
      addr.setPort( std::string(_parent->_parent->ui_lineedit_listenport->text()) );
      Push p(7, 0, id, _qhe.id, _qhe.index, addr);
      _parent->_parent->connections->sendAll( p._data );
      setState(dtPushed);
    }
}


void Download::startPushedDownload(QSocket *socket)
{
  setState(dtConnecting);
  sock = socket;
  
  connect(sock, SIGNAL(readyRead()), this, SLOT(slotReadyRead()));
  connect(sock, SIGNAL(connectionClosed()), this, SLOT(slotConnectionClosed()));
  connect(sock, SIGNAL(error(int)), this, SLOT(slotError(int)));
  
  slotConnected();
}


const int Download::readState()
{
  return _state;
}


void Download::setState(int i)
{
  if(_c->_search_item_set.find(_search_item) == _c->_search_item_set.end()) _search_item = NULL;
  
  switch(i)
    {
    case dtExists:
    case dtError:
    case dtAbort:
    case dtClosed:
    case dtClosedUnfinished:
    case dtNotFound:
    case dtTimeout:
    case dtBusy:
      if(_search_item) _search_item->setPixmap(2, *(Icon::getInstance()->getRedBall()));
      _c->_interrupted_downloads->setRed(_int_item);
      break;

    case dtConnected:
      if(_search_item) _search_item->setPixmap(2, *(Icon::getInstance()->getGreenBall()));
      _c->_interrupted_downloads->setGreen(_int_item);
      break;

    case dtConnecting:
    case dtQueued:
    case dtPushed:
    case dtIdle:
      if(_search_item) _search_item->setPixmap(2, *(Icon::getInstance()->getYellowBall()));
      _c->_interrupted_downloads->setYellow(_int_item);
      break;

    case dtClosedCompleted:
      if(_search_item) _search_item->setPixmap(2, *(Icon::getInstance()->getSmile()));
      _c->_interrupted_downloads->setSmile(_int_item);
      break;
    }

  std::strstream str;
  str << "Set state " << i << "." << std::ends;
  Logging::getInstance()->log(Logging::Debug, str.str());
  str.freeze(false);

  _state = i;
}


const unsigned long Download::bytesRead()
{
  if(_f)
    if(_f->exists())
      return _f->size();

  return 0;
}


const unsigned long Download::bytesToRead()
{
  return _content_length - bytesRead();
}


void Download::slotReadyRead()
{
  int	v,w;
  char  cbuff[32768];

  w = sock->readBlock(cbuff, 32768);
  _buffer.append(cbuff, w);
  tmp_read += w;
  _all_read += w;

  if(_header.empty())
    {
      std::string::size_type  pos = _buffer.find("\r\n\r\n");
      
      if( pos != std::string::npos)
	{
	  _header = _buffer.substr(0, pos+4);
	  _buffer = _buffer.substr(pos+4);

	  std::string::size_type  idx1 = _header.find(" ");
	  std::string::size_type  idx2 = _header.find(" ", idx1+1);

	  if( _header.substr(idx1+1, idx2-idx1-1) == "404" ) 
	    {
	      setState(dtNotFound);
	      return;
	    }
	  
	  if( _header.substr(idx1+1, idx2-idx1-1) == "503" ) 
	    {
	      setState(dtBusy);
	      timer_retry->start(1000);
	      return;
	    }
	  
	  if( _header.substr(idx1+1, idx2-idx1-1) == "200" ) 
	    {
	      setState(dtConnected);
	      timer_bandwidth->start(4000);
	      bandwidth = 0;
	    }

	  // get server name
	  
	  StringManipulation sm(_header);
	  if( ((idx1 = sm.to_lower().find("server:")) != std::string::npos) && (_client.empty()) )
	    {
	      idx1 += 7;  // point to character after ':'
	      idx1 = _header.find_first_not_of(" ", idx1);
	      idx2 = _header.find("\r\n", idx1);
	      _client =  _header.substr(idx1, idx2-idx1);
	    }
	  
	  // save time
	  _start_time = time(NULL);
	}
    }

  if( (!_buffer.empty()) && (readState()==dtConnected) )
    {
	v = _f->size();
	w = _f->writeBlock(_buffer.data(), _buffer.size());
	_f->flush();

	std::strstream str;
	str << "Write bytes: " << v << ", " << _content_length << ", " << w << ", " << _buffer.size() << std::ends;
	Logging::getInstance()->log(Logging::Debug, str.str());
	str.freeze(false);

	// abort download if the write was not ok (e.g. disk full)
	if( ( w != _buffer.size() ) || ( v+w != _f->size()) )  
	  {
	    std::string error("Cannot write to disk [" + std::string(_f->name()) + "]");
	    Logging::getInstance()->log(Logging::Error, error);

	    abortDownload();
	    return;
	  }

	_parent->_parent->connections->_statistic.loads_dsize += w;
	_buffer.erase(0, w);
   }
  
  std::strstream str;
  str << "bytes read: " << bytesRead() << "/" << _all_read << " header: " << _header.size() << std::ends;
  Logging::getInstance()->log(Logging::Debug, str.str());
  str.freeze(false);
  
  if(bytesRead() >= _content_length) finish();
}


void Download::finish()
{
  int loglevel = Logging::Info;
  std::strstream log;
  log << "Download finished. ";
  
  sock->close();
  setState(dtClosed);
  _active   = false;
  _end_time = time(NULL);

  if(readState() != dtBusy)
    if(bytesRead() == _content_length)
      {
	if(_search_item) _search_item->setPixmap(2, *(Icon::getInstance()->getSmile()));
	_c->_interrupted_downloads->setSmile(_int_item);

	log << "[all bytes read]";

	_parent->_parent->connections->_statistic.loads_dsucc++;
	setState(dtClosedCompleted);
	
	_f->close();

	std::string destination = complete_path;
	int    i = 0;

	if(_parent->_parent->ui_radiobutton_e_rename->isChecked() || _parent->_parent->ui_radiobutton_e_abort->isChecked())
	  {
	    QFile d(destination.c_str());

	    std::string di = complete_path.substr(0, complete_path.find_last_of("/"));
	    std::string fi = complete_path.substr(complete_path.find_last_of("/")+1);

	    while(d.exists())
	      {
		destination = (di + "/");
		
		std::strstream str;
		str << i++ << std::ends;
		destination += str.str();
		str.freeze(false);
		
		destination += "_" + fi;

		d.setName(destination.c_str());
	      }
	  }

	QFile dest(destination.c_str());

	QDir d;
	if( !d.rename(_f->name(), destination.c_str()) )
	  {
	    if( ! copyFile(std::string(_f->name()), destination) )
	      {
		if(_search_item) _search_item->setPixmap(2, *(Icon::getInstance()->getGreenSmile()));
		_c->_interrupted_downloads->setGreenSmile(_int_item);

		log << " [error moving file from incomleted path to completed path ("
		    << _f->name() << " -> " << destination << ")]";
		loglevel = Logging::Error;
		_filename = _f->name();
	      }
	    else
	      _f->setName(destination.c_str());
	  }
	else
	  {
	    _f->setName(destination.c_str());
	    _filename = destination;
	  }
      }
    else
      {
	if(_search_item) _search_item->setPixmap(2, *(Icon::getInstance()->getYellowBall()));
	_c->_interrupted_downloads->setYellow(_int_item);
	    
	log << "[bytes read (" << bytesRead() << ")" << " != content length (" << _content_length << ")]";
	loglevel = Logging::Error;
	_parent->_parent->connections->_statistic.loads_dabor++;
	setState(dtClosed);
	timer_retry->start(1000);
      }
  else
    {
      log << "[state is busy]";
      loglevel = Logging::Error;
    }

  log << std::ends;
  Logging::getInstance()->log(loglevel, log.str());
  log.freeze(false);
}


void Download::slotConnectionClosed()
{
  std::strstream log;
  log << "Connection closed. Socket buffer: " << sock->bytesAvailable() 
      << " Buffer: " << _buffer.size() << " State: " << readState() 
      << " Filesize: " << _f->size() << std::endl;

  while(sock->bytesAvailable() > 0) slotReadyRead();

  if(readState() == dtConnected) _f->writeBlock(_buffer.data(), _buffer.size());
  _buffer.erase();
  _f->flush();

  log << "Flushed." << std::endl
      << "Socket buffer: " << sock->bytesAvailable()
      << " Buffer: " << _buffer.size() << " State: " << readState()
      << " Filesize: " << _f->size() << std::ends;

  Logging::getInstance()->log(Logging::Debug, log.str());
  log.freeze(false);

  finish();

  _end_time = time(NULL);
}


void Download::slotError(int error)
{
  std::strstream log;
  log << "Download error. [code=" << error << "]";
  Logging::getInstance()->log(Logging::Error, log.str());
  log.freeze(false);

  setState(dtError);
  timer_retry->start(1000);
}


void Download::slotConnected()
{
  std::string     s = _qhe.filename;
  unsigned long   begin = _f->size();
  std::strstream  str;

  str << "GET /get/" << _qhe.index << "/" << s << " HTTP/1.0\r\n"
      << "Range: bytes=" << begin << "-\r\n"
      << "User-Agent: Qtella " << VERSION << "\r\n\r\n" << std::ends;

  _start_read = begin;

  s = str.str();
  str.freeze(false);

  sock->writeBlock(s.c_str(), s.size());
}



void Download::slotTimerBandwidth()
{
  bandwidth = static_cast<double>(tmp_read) / 4096;  // KB/s
  tmp_read = 0;
}


const double Download::readBandwidth()
{
  return bandwidth;
}


void Download::slotTimeout()
{
  timeout->stop();

  if(readState() == dtConnecting)
    {
      setState(dtTimeout);
      sock->close();
      timer_retry->start(1000);
    }
}


void Download::abortDownload()
{
  Logging::getInstance()->log(Logging::Debug, "Abort download.");

  timeout->stop();
  timer_bandwidth->stop();
  timer_retry->stop();
  if(sock) sock->close();
  
  if(_f) _f->close();
  
  _parent->_parent->connections->_statistic.loads_dabor++;
  setState(dtAbort);
  _active = false;
}


void Download::resumeDownload()
{
  _retries = 0;
  setState(dtQueued);
}


Download::~Download()
{
  if(_c->_search_item_set.find(_search_item) == _c->_search_item_set.end()) _search_item = NULL;

  if(_search_item) _search_item->setPixmap(2, QPixmap());
  _c->_interrupted_downloads->clearColored(_int_item);

  delete timeout;
  delete timer_bandwidth;
  delete timer_retry;
  if(sock) delete sock;
  delete _item;

  if(_f)
    {
      if(_f->isOpen()) _f->close();
      if(_f->exists()) if(_f->size() == 0) _f->remove();
      delete _f;
    }
}

void Download::slotRetry()
{
  retry_timeout--;

  if( retry_timeout == 0 )  
    {
      if(timer_retry->isActive()) timer_retry->stop();
      if(timeout->isActive()) timeout->stop();
      if(timer_bandwidth->isActive()) timer_bandwidth->stop();

      setState(dtQueued);
      _retries++;

      int n = _parent->_parent->ui_spinbox_conf_retries->value();
      if( (n != -1) && (_retries > n) ) 
	{
	  setState(dtError);
	  _active = false;
	}
    }
}


const int Download::readRetryTimeout()
{
  return retry_timeout;
}


std::string Download::createFilename(std::string file, unsigned long size, std::string host)
{
  time_t t  = time(NULL);

  for(int id = rand()%1000; ; ++id)
    {
      std::strstream str;
      std::string    s;

      std::string::size_type  idx = file.find_last_of(".");
      std::string             ext;
      std::string             filename = file;
      if(idx != std::string::npos)
	{
	  filename = file.substr(0, idx);
	  ext = file.substr(idx);  // with "." !!
	}

      str << ".part_" << t << "_" << size << "_" << host << "_" << id << std::ends;
      s = filename + std::string(str.str()) + ext;
      str.freeze(false);

      QFile f(s.c_str());

      if(!f.exists()) return s;
    }
}


std::string Download::findFile(std::string file, unsigned long size)
{
  std::string::size_type  idx = file.find_last_of("/");

  std::string    directory = file.substr(0, idx);
  std::string    filename  = file.substr(idx+1);
  unsigned long  current_size = 0;

  std::string            ext;
  std::string            name;
  std::string::size_type p;
  if( (p = filename.find_last_of(".")) != std::string::npos)
    {
      ext = filename.substr(p+1);
      name = filename.substr(0, p);
    }

  QDir d(directory.c_str(), "*", QDir::DefaultSort, QDir::Files);

  std::string s;

  for(int i = 0; i < d.count(); ++i)
    {
      filename = d[i];

      if( (idx = filename.find(".part_")) == std::string::npos ) continue;
      if( (filename.substr(0, idx) != name) || (ext != filename.substr(filename.find_last_of(".")+1)) ) continue;

      idx = filename.find("_", idx+6);  // find first "_"
      std::string::size_type  idx2 = filename.find("_", idx+1); // find second "_" 

      std::string s_size = filename.substr(idx+1, idx2-idx-1);  // extract filesize
      
      std::strstream str;
      str << s_size << std::ends;
      unsigned long str_size;
      str >> str_size;
      
      std::string s_new = directory + "/" + filename;
      QFile f(s_new.c_str());
      
      if( (str_size == size) && (f.size() > current_size) )
	{
	  s = s_new;
	  current_size = f.size();
	}
    }

  return s;
}


std::string Download::findFile(std::string file, std::string host)
{
  std::string::size_type  idx = file.find_last_of("/");

  string         directory = file.substr(0, idx);
  string         filename  = file.substr(idx+1);
  unsigned long  current_size = 0;

  std::string            ext;
  std::string            name;
  std::string::size_type p;
  if( (p = filename.find_last_of(".")) != std::string::npos)
    {
      ext = filename.substr(p+1);
      name = filename.substr(0, p);
    }

  QDir d(directory.c_str(), "*", QDir::DefaultSort, QDir::Files);

  std::string s;

  for(int i = 0; i < d.count(); ++i)
    {
      filename = d[i];

      if( (idx = filename.find(".part_")) == std::string::npos ) continue;
      if( (filename.substr(0, idx) != name) || (ext != filename.substr(filename.find_last_of(".")+1)) ) continue;

      idx = filename.find("_", idx+6);
      idx = filename.find("_", idx+1);
      std::string::size_type  idx2 = filename.find("_", idx+1);
      
      std::string s_host = filename.substr(idx+1, idx2-idx-1);
      
      if( (s_host == host) )
	{
	  std::string s_new = directory + "/" + filename;
	  QFile f(s_new.c_str());
	  
	  if( f.size() > current_size )
	    {
	      s = s_new;
	      current_size = f.size();
	    }
	}
    }

  return s;
}


std::string Download::findFile(std::string file, unsigned long size, std::string host)
{
  std::string::size_type  idx = file.find_last_of("/");

  std::string         directory = file.substr(0, idx);
  std::string         filename  = file.substr(idx+1);
  unsigned long  current_size = 0;

  std::string            ext;
  std::string            name;
  std::string::size_type p;
  if( (p = filename.find_last_of(".")) != std::string::npos)
    {
      ext = filename.substr(p+1);
      name = filename.substr(0, p);
    }

  QDir d(directory.c_str(), "*", QDir::DefaultSort, QDir::Files);

  std::string s;

  for(int i = 0; i < d.count(); ++i)
    {
      filename = d[i];

      if( (idx = filename.find(".part_")) == std::string::npos ) continue;
      if( (filename.substr(0, idx) != name) || (ext != filename.substr(filename.find_last_of(".")+1)) ) continue;

      idx  = filename.find("_", idx+6);

      std::string::size_type  idx2;
      std::string::size_type  idx3;
      
      idx2 = filename.find("_", idx+1);
      idx3 = filename.find("_", idx2+1);
      
      std::string s_size = filename.substr(idx+1, idx2-idx-1);
      std::string s_host = filename.substr(idx2+1, idx3-idx2-1);
      
      std::strstream str;
      str << s_size << std::ends;
      unsigned long str_size;
      str >> str_size;
      
      if( (s_host == host) && (str_size == size) )
	{
	  std::string s_new = directory + "/" + filename;
	  QFile f(s_new.c_str());
	  
	  if( f.size() > current_size )
	    {
	      s = s_new;
	      current_size = f.size();
	    }
	}
    }

  return s;
}


void Download::slotUpdateStatus()
{
  static const std::string status[] = 
  {
    "Idle", "Error", "Abort", "Closed", "Completed", "Closed",
    "Connecting", "NotFound", "Timeout", "Busy", "Connected", "Exist",
    "Queued", "Push" 
  };

  std::string s[8];
  for(int i = 0; i < 8; ++i) s[i] = "";

  s[1] = status[readState()];

  if( (readState()==dtBusy) || (readState()==dtClosed) || (readState()==dtTimeout) || 
      ( (readState()==dtError) && _active ) )
    {
      s[1] += " - ";
      std::strstream str;
      str << readRetryTimeout() << "s" << std::ends;
      s[1] += str.str();
      str.freeze(false);
    }

  if( readState() == dtClosedCompleted )
    {
      time_t         diff = _end_time - _start_time;
      std::strstream str;
      int            h   = diff / 3600;
      int            m   = (diff - h * 3600) / 60;
      int            sec = diff - h * 3600 - m * 60;
      str.fill('0');
      str << h << ":" << std::setw(2) << m << ":" << std::setw(2) << sec << std::ends;
      s[6].assign(std::string(str.str()));
      str.freeze(false);

      if(diff>0)
	{
	  double b = (double)(bytesRead() - _start_read) / diff / 1024;
	  std::strstream str;
	  str.setf(ios::fixed);
	  str << std::setprecision(2) << b << " KB/s" << std::ends;
	  s[5] = str.str();
	  str.freeze(false);
	}
    }

  if( readState() == dtConnected )
    {
      // bandwidth
      std::strstream str;
      str.setf(ios::fixed);
      str << std::setprecision(2) << readBandwidth() << " KB/s" << std::ends;
      s[5] = str.str();
      str.freeze(false);

      // calculate time to go
      uint32_t seconds   = 0;
      uint32_t toread    = bytesToRead();
      uint32_t read      = bytesRead() - _start_read;
      double bandwidth   = readBandwidth();
      if(read > 0) 
	{
	  double d = static_cast<double>(toread) / static_cast<double>(read);
	  d *= static_cast<double>(time(NULL) - _start_time);
	  seconds = static_cast<uint32_t>(d);
	}
      int      h   = seconds / 3600;
      int      m   = (seconds - h * 3600) / 60;
      int      sec = seconds - h * 3600 - m * 60;
      std::strstream time_str;
      time_str.fill('0');
      time_str << h << ":" << std::setw(2) << m << ":" << std::setw(2) << sec << std::ends;
      s[6] = time_str.str();
      time_str.freeze(false);
    }

  // progress
  double p = 100;
  if(_content_length > 0) p = static_cast<double>(100) * static_cast<double>(bytesRead()) / static_cast<double>(_content_length);
  std::strstream str;
  str.setf(std::ios::fixed);
  str << std::setprecision(1) << p << "%" << std::ends;
  s[3] = str.str();
  str.freeze(false);

  s[0] = _qhe.filename;
  if(_dest_file != _qhe.filename) s[0] += " [saved as: " + _dest_file + "]";

  s[4] = _qhe.addr.strIP();
  s[7] = _client;

  str.seekp(0);
  str << _qhe.size << std::ends;
  s[2] = StringManipulation::insertDelimiter(str.str(), ',', 3);
  str.freeze(false);

  for(int i = 0; i < 8; ++i) _item->setText(i, s[i].c_str());
}


bool Download::isActive()
{
  return _active;
}


bool Download::moveFile(std::string directory)
{
  std::string current(_filename);
  if(current.find_last_of("/") != std::string::npos) current.erase(0, current.find_last_of("/")+1);

  QDir d;
  if( d.rename(_filename.c_str(), (directory + current).c_str()) )
    {
      _f->setName( (directory+current).c_str() );
      _filename = directory + current;
    }
  else 
    if( copyFile(_filename, directory + current) )
      {
	_f->setName( (directory+current).c_str() );
	_filename = directory + current;
      }
    else return false;

  return true;
}


std::string Download::getFileName()
{
  return _filename;
}


bool Download::copyFile(std::string src, std::string dst)
{
  QFile fsrc(src.c_str());
  QFile fdst(dst.c_str());

  if( ! fsrc.open( IO_ReadOnly ) ) return false;
  if( ! fdst.open( IO_WriteOnly ) ) return false;

  int  r = 0;
  char buffer[4096];

  while( ( r = fsrc.readBlock(buffer, 4096) ) > 0 )
    if( r != fdst.writeBlock(buffer, r) ) return false;

  fsrc.close();
  fsrc.remove();

  _filename = dst;
  return true;
}
